#include "fastlz.h"
#include "filesystem.h"
#include "mm_types.h"
#include "nds/arm9/background.h"
#include "nds/arm9/cache.h"
#include "nds/arm9/console.h"
#include "nds/arm9/input.h"
#include "nds/arm9/sassert.h"
#include "nds/cothread.h"
#include "nds/decompress.h"
#include "nds/dma.h"
#include "nds/interrupts.h"
#include "nds/timers.h"
#include "raylibds.hpp"
#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <nds.h>
#include <maxmod9.h>
#include <sys/_types.h>

#define PALETTE_SIZE 256
#define FRAME_SIZE 256*192

// TODO: hardcode fix
#define CHUNK_SIZE 8
#define FRAMERATE 24
#define VIDEO_TOTALFRAME 3268

#define VIDEO_RBUF_LEN CHUNK_SIZE*8
#define VIDEO_RBUF_BUFFER FRAME_SIZE
#define VIDEO_TEMP_SIZE 204000

#define MMSTREAM_BUF_SIZE 800
#define MUSIC_RBUF_SIZE 96000*2

void wait_forever(const char* msg);
bool LoadNextChunk(unsigned char *dest);
void TimerCallback();
void VBLCallback();
void FrameStep();
int ThreadMusic(void *arg);
int ThreadVideo(void *arg);
mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format );

volatile bool doDraw = true;
volatile bool isPause = false;
volatile int noframe = 0;

void *sassert_malloc(size_t size)
{
	void *p = malloc(size);
	sassert(p != nullptr, "Memory allocation failed %i", size);
	return p;
}

struct RingBuffer {
	unsigned char *buffer = nullptr;
	int head;
	int tail;
	size_t buffer_size;
	size_t size;
	size_t count;

	RingBuffer(size_t rb_len, size_t buffer_len, unsigned char * buffer_ptr)
	{
		buffer = buffer_ptr;
		buffer_size = buffer_len;
		size = rb_len;
		head = 0;
		tail = 0;
		count = 0;
		Raylib::nocashMessageFormat("per block %i", buffer_size/size);
	}

	unsigned char* Get()
	{
		if (count > 0)
		{
			count--;
			unsigned char* ptr = &buffer[head*(buffer_size/size)];
			head = (head+1) % size;
			return ptr;
		}

		return nullptr;
	}

	unsigned char* Write()
	{
		if (count < size)
		{
			count++;
			unsigned char* ptr = &buffer[tail*(buffer_size/size)];
			tail = (tail+1) % size;
			return ptr;
		}

		return nullptr;
	}

	unsigned char* Manual_Read()
	{
		unsigned char* ptr = &buffer[head*(buffer_size/size)];
		return ptr;
	}

	unsigned char* Manual_Write()
	{
		unsigned char* ptr = &buffer[tail*(buffer_size/size)];
		return ptr;
	}

	void Manual_Skip(int total)
	{
		count += total;
		tail = (tail+total) % size;
	}

	void Manual_Advance(int total)
	{
		count -= total;
		head = (head+total) % size;
	}

	size_t getFree()
	{
		return size-count;
	}
};

struct Rbuf_Music {
	unsigned char *buffer = nullptr;
	int head;
	int tail;
	int size;
	int count;

	Rbuf_Music(size_t buffer_size, unsigned char* buffer_ptr)
	{
		buffer = buffer_ptr;
		size = buffer_size;
		head = 0;
		tail = 0;
		count = 0;
	}

	void Skip(int total)
	{
		count -= total;
		head = (head+total) % size;
	}

	void Advance(int total)
	{
		count += total;
		tail = (tail+total) % size;
	}

	int LinearReadSize()
	{
		return size - head;
	}

	int LinearWriteSize()
	{
		return size - tail;
	}

	size_t GetFree()
	{
		return size-count;
	}

	unsigned char *GetHead()
	{
		return &buffer[head];
	}

	unsigned char *GetTail()
	{
		return &buffer[tail];
	}

	int read(char *dest, size_t len)
	{
		if (count >= len)
		{
			count -= len;
			int linear_len = size - head;

			if (len < linear_len)
				memcpy(dest, &buffer[head], len);
			else
			{
				memcpy(dest, &buffer[head], linear_len);
				memcpy(&dest[linear_len], buffer, len-linear_len);
			}

			head = (head+len) % size;

			return 1;
		}

		return 0;
	}

	int write(char *data, size_t len)
	{
		if (count+len <= size)
		{
			count += len;
			int linear_len = size - tail;

			if (len < linear_len)
				memcpy(&buffer[tail], data, len);
			else
			{
				memcpy(&buffer[tail], data, linear_len);
				memcpy(buffer, &data[linear_len], len-linear_len);
			}

			tail = (tail+len) % size;

			return 1;
		}

		return 0;
	}
};

int ptr_background;
int ptr_subbackground;

FILE *file_music = nullptr;
int file_music_len;

FILE *file_videomain = nullptr;
int file_videomain_len;

RingBuffer *videomain_rb;
unsigned char *frame_preload_temp;
unsigned char *frame_read_temp;

Rbuf_Music *music_rb;
unsigned char *music_buffer_temp;

uint32_t preload_video_size = 0;
uint32_t next_video_size = 0;

volatile int timer_start = 0;
volatile int timer_avg = 0;

int main(void) {

	videoSetMode(MODE_5_3D);
	videoSetModeSub(MODE_5_2D);

	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankC(VRAM_C_SUB_BG);
	vramSetBankF(VRAM_F_TEX_PALETTE);

	consoleDemoInit();
	consoleDebugInit(DebugDevice_NOCASH);

	bool nitrofs = nitroFSInit(NULL);
	sassert(nitrofs, "nitrofs init failed");

	frame_preload_temp = (unsigned char*)sassert_malloc(VIDEO_TEMP_SIZE);
	music_buffer_temp = (unsigned char*)sassert_malloc(MUSIC_RBUF_SIZE);

	videomain_rb = new RingBuffer(VIDEO_RBUF_LEN, VIDEO_RBUF_LEN*VIDEO_RBUF_BUFFER, (unsigned char*)sassert_malloc(VIDEO_RBUF_LEN*VIDEO_RBUF_BUFFER));
	music_rb = new Rbuf_Music(MUSIC_RBUF_SIZE, (unsigned char*)sassert_malloc(MUSIC_RBUF_SIZE));

	memset(music_rb->buffer, 0, MUSIC_RBUF_SIZE);

	file_videomain = fopen("nitro:/main_img.bin", "rb");
	sassert(file_videomain != nullptr, "failed to load image.bin");
	fseek(file_videomain, 0, SEEK_END);
	file_videomain_len = ftell(file_videomain);
	fseek(file_videomain, 0, SEEK_SET);

	file_music = fopen("nitro:/music1.raw", "rb");
	sassert(file_music != nullptr, "failed to load music.raw");
	fseek(file_music, 0, SEEK_END);
	file_music_len = ftell(file_music);
	fseek(file_music, 0, SEEK_SET);

	ptr_background = bgInit(3, BgType_Bmp8, BgSize_B8_256x256, 0, 0);
	// ptr_subbackground = bgInitSub(3, BgType_Bmp8, BgSize_B8_256x256, VRAM_C_SUB_BG, 0);

	int palmain_len;
	unsigned char* palmain = Raylib::LoadFile("nitro:/main_pal.bin", palmain_len);

	DC_FlushRange(palmain, palmain_len);
	dmaCopy(palmain, BG_PALETTE, palmain_len);

	Raylib::UnloadFile(palmain);

	{
		Raylib::nocashMessageFormat("preload video main");
		do {
			unsigned char *ptr = videomain_rb->Manual_Write();
			LoadNextChunk(ptr);
			videomain_rb->Manual_Skip(CHUNK_SIZE);
		} while(videomain_rb->count+CHUNK_SIZE < videomain_rb->size);

		Raylib::nocashMessageFormat("preload done");
	}

	{
		Raylib::nocashMessageFormat("preload audio");
		int len = music_rb->GetFree();
		int adv = fread(music_rb->GetTail(), sizeof(char), len, file_music);
		music_rb->Advance(adv);
		Raylib::nocashMessageFormat("preload done");
	}

	mm_ds_system sys;
	sys.mod_count 			= 0;
	sys.samp_count			= 0;
	sys.mem_bank			  = 0;
	sys.fifo_channel		= FIFO_MAXMOD;
	mmInit(&sys);

	mm_stream mystream;
	mystream.sampling_rate	= 22050;
	mystream.buffer_length	= MMSTREAM_BUF_SIZE;
	mystream.callback		= on_stream_request;
	mystream.format			= MM_STREAM_16BIT_STEREO;
	mystream.timer			= MM_TIMER1;
	mystream.manual			= false;

	timerStart(0, ClockDivider_1024, TIMER_FREQ_1024(FRAMERATE), TimerCallback);
	irqSet(IRQ_VBLANK, []{
		printf("\033[4;2H\033[Kvideo buffer: %i frames\n", videomain_rb->count);
		printf("\033[5;2H\033[Kmusic buffer: %i bytes\n", music_rb->count);
		if (noframe) printf("\033[6;2H\033[Kno frame: %i\n", noframe);

		// static int counter = 0;
		// counter++;

		// if (counter % 30 == 0)
		// 	printf("\033[6;2H\033[Kavg decode time: %ims\n", timerTicks2msec(timer_avg/30));

		scanKeys();
		if (keysUp() & KEY_A) isPause = !isPause;
		if (keysUp() & KEY_B) FrameStep();
		if (keysHeld() & KEY_Y) FrameStep();
	});

	cothread_create(ThreadMusic, NULL, 0, COTHREAD_DETACHED);
	// cothread_create(ThreadVideo, NULL, 0, COTHREAD_DETACHED);

	// cpuStartTiming(1);

	mmStreamOpen( &mystream );

	while(1) {
		cothread_yield();
		// swiWaitForVBlank();
		// cothread_yield_irq(IRQ_VBLANK);
		// scanKeys();


		// if (keysUp() & KEY_A) FrameStep();

		if (!isPause)
		{
			if (file_videomain)
			{
				unsigned char *ptr = videomain_rb->Manual_Write();
				if (videomain_rb->count+CHUNK_SIZE < videomain_rb->size)
				{
					// timer_start = cpuGetTiming();
					LoadNextChunk(ptr);
					videomain_rb->Manual_Skip(CHUNK_SIZE);
					// timer_avg = cpuGetTiming() - timer_start;
				}
			}

			if (ftell(file_videomain) == file_videomain_len)
			{
				fclose(file_videomain);
				file_videomain = nullptr;
				// return 0;
			}
		}

	}

	return 0;
}

int ThreadVideo(void *arg)
{
	// while(1) {
	// 	if (videomain_rb->getFree() > (sizeof(uint32_t)+preload_video_size) && videomain_rb->getFree() > videomain_rb->size/2)
	// 	{
	// 		// Raylib::nocashMessageFormat("video thread %i %i", videomain_rb->getFree(), preload_video_size+4);

	// 		// if (comutex_try_acquire(&mutex))
	// 		// {
	// 			bool preload;
	// 			do {
	// 				preload = LoadNextChunk(file_videomain);
	// 			} while(preload);
	// 			// comutex_release(&mutex);
	// 		// }
	// 	}

	// 	if (ftell(file_videomain) == file_videomain_len)
	// 	{
	// 		fclose(file_videomain);
	// 		file_videomain = nullptr;
	// 		return 0;
	// 	}
	// 	cothread_yield();
	// }

	// // Raylib::nocashMessageFormat("video thread exit");
	// return 0;
}

int ThreadMusic(void *arg)
{
	while(1) {
		cothread_yield();

		if (file_music != nullptr && music_rb->count < music_rb->size*0.35)
		{
			int free = music_rb->GetFree();
			int linear = music_rb->LinearWriteSize();
			int adv = 0;

			// memset(music_buffer_temp, 0, MUSIC_RBUF_SIZE);
			// fread(music_buffer_temp, free, 1, file_music);


			if (free < linear)
			{
				// memcpy(music_rb->GetTail(), music_buffer_temp, free);
				adv += fread(music_rb->GetTail(), 1, free, file_music);
			}
			else
			{
				// memcpy(music_rb->GetTail(), music_buffer_temp, linear);
				// memcpy(music_rb->buffer, music_buffer_temp, free-linear);
				adv += fread(music_rb->GetTail(), 1, linear, file_music);
				adv += fread(music_rb->buffer, 1, free-linear, file_music);
			}

			music_rb->Advance(adv);

		}

		if (ftell(file_music) >= file_music_len)
		{
			fclose(file_music);
			file_music = nullptr;
			Raylib::nocashMessageFormat("music close!");
		}

		static bool closeOnce = false;

		if (!closeOnce && file_music == nullptr && music_rb->count <= 0)
		{
			mmStreamClose();
			Raylib::nocashMessageFormat("stream close!");
			closeOnce = true;
		}
	}
	
	return 0;
}

void FrameStep()
{
	if (videomain_rb->count == 0 && !file_videomain) return;
	
	if (videomain_rb->count)
	{
		// if (comutex_try_acquire(&mutex))
		// {
			// videomain_rb->Get();
			// videomain_rb->read((unsigned char*)&next_video_size, sizeof(uint32_t));
			unsigned char *dest = videomain_rb->Get();
			// Raylib::nocashMessageFormat("frame step start %p %i", dest, videomain_rb->count);

			// videomain_rb->read(frame_read_temp, next_video_size);

			// DC_FlushRange(dest, VIDEO_RBUF_BUFFER);
			// decompress(dest, bgGetGfxPtr(ptr_background), DecompressType::LZ77Vram);
			dmaCopyAsynch(dest, bgGetGfxPtr(ptr_background), FRAME_SIZE);
			// dmaCopyAsynch(frame_temp, bgGetGfxPtr(ptr_background), FRAME_SIZE);

			// comutex_release(&mutex);
			// Raylib::nocashMessageFormat("frame step end");
		} else {
			printf("no frame!!!!!\n");
			noframe++;
		}
	// }
}

void TimerCallback()
{
	FrameStep();
	doDraw = false;

	// printf("\033[4;2H\033[Kvideo buffer: %i", videomain_rb->count/(FRAME_SIZE));
	// printf("\033[5;2H\033[Kmusic buffer: %i", music_rb->count/(MMSTREAM_BUF_SIZE*4));
}

bool LoadNextChunk(unsigned char *dest)
{
	if (!dest) return false;
	// Raylib::nocashMessageFormat("load next start, %p", dest);

	// videomain_rb->write((unsigned char*)&preload_video_size, sizeof(uint32_t));
	
	fread(&preload_video_size, sizeof(uint32_t), 1, file_videomain);
	fread(frame_preload_temp, sizeof(uint8_t), preload_video_size, file_videomain);
	decompress(frame_preload_temp, dest, DecompressType::LZ77);
	// fastlz_decompress(frame_preload_temp, preload_video_size, dest, FRAME_SIZE*CHUNK_SIZE);

	return true;
}

mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format ) {
	size_t samplesize = 1;
	switch(format){
		case MM_STREAM_8BIT_MONO: samplesize = 1; break;
		case MM_STREAM_8BIT_STEREO: samplesize = 2; break;
		case MM_STREAM_16BIT_MONO: samplesize = 2; break;
		case MM_STREAM_16BIT_STEREO: samplesize = 4; break;
	}
	int len = length*samplesize;
	int linear = music_rb->LinearReadSize();

	// Raylib::nocashMessageFormat("%i", len);
	
	// if (ftell(file_music) >= file_music_len) mmStreamClose();
	// fread(music_buffer_temp, 1, len, file_music);
	// memcpy(dest, music_buffer_temp, len);
	// return length;

	// if (len < linear)
	// {
	// 	memcpy(dest, music_rb->GetHead(), len);
	// 	music_rb->Skip(len);
	// 	return length;
	// }
	// else
	// {
	// 	// Raylib::nocashMessageFormat("linear %i", linear);
	// 	memcpy(dest, music_rb->GetHead(), linear);
	// 	// memcpy((char*)dest+linear, music_rb->buffer, len-linear);
	// 	music_rb->Skip(linear);
	// 	return linear/samplesize;
	// }


	if (len < linear)
	{
		memcpy(dest, music_rb->GetHead(), len);
	}
	else
	{
		// Raylib::nocashMessageFormat("linear %i", linear);
		memcpy(dest, music_rb->GetHead(), linear);
		memcpy((char*)dest+linear, music_rb->buffer, len-linear);
	}

	music_rb->Skip(len);
	return length;
}